#!/usr/bin/python
# -*- encoding: utf-8 -*-
import inspect
import os
import sys
from types import ClassType, FunctionType, MethodType
from log import Logger


class TraceContext (object):
    def __init__(self, output=lambda s: sys.stderr.write(s + '\n'), logfile=None):
        self._output = output
        self._depth = 0
        self._inc = 0
        self.log = Logger('pydev', logfile=logfile)

    def __call__(self, thing):
        t = type(thing)
        if t is FunctionType:
            return self._wrap_funclike(thing)
        elif t is MethodType:
            return self._wrap_function(thing)
        elif t in [type, ClassType]:
            return self._wrap_type(thing)
        else:
            raise NotImplementedError('Cannot wrap: {0!r}'.format(t))

    def info(self, msg):
        prefix = '| ' * self._depth
        msg = '{0}[I] {1}'.format(prefix, msg)
        self.log.info(msg)

    def debug(self, msg):
        prefix = '| ' * self._depth
        msg = '{0}[D] {1}'.format(prefix, msg)
        self.log.debug(msg)

    def error(self, msg):
        Logger('yu')
        prefix = '| ' * self._depth
        msg = '{0}[E] {1}'.format(prefix, msg)
        self.log.warning(msg)

    def warning(self, msg):
        prefix = '| ' * self._depth
        msg = '{0}[W] {1}'.format(prefix, msg)
        self.log.warning(msg)

    # Private:
    def _log(self, tmpl, *args, **kw):
        prefix = '| ' * self._depth

        msg = '{0}{1}'.format(prefix, tmpl.format(*args, **kw))

        self.log.debug(msg)
        #Sself._output(msg)

    def _wrap_funclike(self, c):
        (argnames, _, _, _) = inspect.getargspec(c)
        if argnames and argnames[0] == 'self':
            return self._wrap_method(c)
        else:
            return self._wrap_function(c)

    def _wrap_method(self, m):
        def wrapped(wrappedSelf, *args, **kw):
            name = '{0}.{1}'.format(wrappedSelf.__class__.__name__, m.__name__)
            fullArgs = (wrappedSelf,) + args
            return self._trace_call(name, 1, m, fullArgs, kw)
        return wrapped

    #    def log(self, msg):
    #        return

    def _wrap_function(self, f):
        name = f.__name__
        return lambda *a, **kw: self._trace_call(name, 0, f, a, kw)

    def _wrap_type(self, basetype):
        classdict = {}
        for n, v in vars(basetype).items():
            if type(v) is FunctionType:
                v = self(v)
            classdict[n] = v
        if type(basetype) is ClassType:
            bases = (basetype, object)
        else:
            bases = (basetype,)
        return type(basetype.__name__, bases, classdict)

    def _trace_call(self, name, argStart, f, args, kw):
        file_func = os.path.basename(f.__globals__.get('__file__', '/'))
        file_func = os.path.splitext(file_func)[0]
        self.logger = file_func
        argDesc = self._describeArgs(args[argStart:], kw)
        self._inc += 1
        _inc = self._inc
        self.log.loggername = file_func
        self._log('.-[{0}] > {1}{2}', _inc, name, argDesc)
        try:
            self._depth += 1
            try:
                result = f(*args, **kw)
            finally:
                self._depth -= 1
        except Exception, e:
            self._log('{0} {1}{3} raising {2!r}', _inc, name, argDesc, e)
            raise
        else:
            self._log("'-[{0}] < RETORNO[{3!r}]", _inc, name, argDesc, result)
        return result

    @staticmethod
    def _describeArgs(args, kw):
        fields = [repr(a) for a in args]
        fields.extend(['{0}={1!r}'.format(k, v) for (k, v) in sorted(kw.items())])
        return '({0})'.format(', '.join(fields))


if __name__ == '__main__':
    traced = TraceContext(logfile='/tmp/log.1')
    from functools import wraps

    class TraceCalls(object):
        """ Use as a decorator on functions that should be traced. Several
            functions can be decorated - they will all be indented according
            to their call depth.
        """
        def __init__(self, stream=sys.stdout, indent_step=2, show_ret=False):
            self.stream = stream
            self.indent_step = indent_step
            self.show_ret = show_ret

            # This is a class attribute since we want to share the indentation
            # level between different traced functions, in case they call
            # each other.
            TraceCalls.cur_indent = 0

        def __call__(self, fn):
            @wraps(fn)
            def wrapper(*args, **kwargs):
                indent = ' ' * TraceCalls.cur_indent
                argstr = ', '.join(
                    [repr(a) for a in args] +
                    ["%s=%s" % (a, repr(b)) for a, b in kwargs.items()])
                self.stream.write('%s%s(%s)\n' % (indent, fn.__name__, argstr))

                TraceCalls.cur_indent += self.indent_step
                ret = fn(*args, **kwargs)
                TraceCalls.cur_indent -= self.indent_step
                if self.show_ret:
                    self.stream.write('%s--> %s\n' % (indent, ret))
                return ret
            return wrapper

        #    traced = TraceCalls()

    @traced
    def w(s,y):
        #traced.info('MESSAGE WWW')
        return sum([s,y])

    @traced
    def y(s,y):
        #traced.error('MESSAGE ***')
        w(8,7)
        return s,y

    @traced
    def z(s,y):
        #traced.info('MESSAGE 001')
        w(8,7)
        #traced.debug('MESSAGE')
        return s,y

    w(1,5)
    y(1,5)

    z(1,5)
    traced.info('MSF')